import typing

import lms.util.net

BACKEND_TYPE_CANVAS: str = 'canvas'

BACKEND_TYPES: typing.List[str] = [
    BACKEND_TYPE_CANVAS,
]

OUTPUT_FORMAT_JSON: str = 'json'
OUTPUT_FORMAT_TABLE: str = 'table'
OUTPUT_FORMAT_TEXT: str = 'text'

OUTPUT_FORMATS: typing.List[str] = [
    OUTPUT_FORMAT_JSON,
    OUTPUT_FORMAT_TABLE,
    OUTPUT_FORMAT_TEXT,
]

BACKEND_REQUEST_CLEANING_FUNCS: typing.Dict[str, typing.Callable] = {
    BACKEND_TYPE_CANVAS: lms.util.net.clean_canvas_response,
}



import lms.backend.testing
import lms.model.testdata.users

def test_courses_users_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing course users. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {},
            [
                lms.model.testdata.users.COURSE_USERS['1']['course-admin@test.edulinq.org'],
                lms.model.testdata.users.COURSE_USERS['1']['course-grader@test.edulinq.org'],
                lms.model.testdata.users.COURSE_USERS['1']['course-other@test.edulinq.org'],
                lms.model.testdata.users.COURSE_USERS['1']['course-owner@test.edulinq.org'],
                lms.model.testdata.users.COURSE_USERS['1']['course-student@test.edulinq.org'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_users_list, test_cases)



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model.assignments

BASE_ENDPOINT = "/api/v1/courses/{course_id}/assignments/{assignment_id}"

def request(backend: typing.Any,
        course_id: int,
        assignment_id: int,
        ) -> typing.Union[lms.backend.canvas.model.assignments.Assignment, None]:
    """ Fetch a single course assignment. """

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id, assignment_id = assignment_id)
    headers = backend.get_standard_headers()

    raw_object = lms.backend.canvas.common.make_get_request(url, headers, raise_on_404 = False)
    if (raw_object is None):
        return None

    return lms.backend.canvas.model.assignments.Assignment(**raw_object)



import typing

import lms.backend.canvas.courses.assignments.fetch
import lms.backend.canvas.courses.assignments.list
import lms.backend.canvas.courses.users.fetch
import lms.backend.canvas.courses.users.list
import lms.model.assignments
import lms.model.backend
import lms.model.users
import lms.util.parse

class CanvasBackend(lms.model.backend.APIBackend):
    """ An API backend for Instructure's Canvas LMS. """

    def __init__(self,
            server: str,
            token: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(server, **kwargs)

        if (token is None):
            raise ValueError("Canvas backends require a token.")

        self.token: str = token

    def get_standard_headers(self) -> typing.Dict[str, str]:
        """ Get standard Canvas headers. """

        return {
            "Authorization": f"Bearer {self.token}",
        }

    def courses_assignments_fetch(self,
            course_id: str,
            assignment_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.assignments.Assignment, None]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_assignment_id = lms.util.parse.required_int(assignment_id, 'assignment_id')
        return lms.backend.canvas.courses.assignments.fetch.request(self, parsed_course_id, parsed_assignment_id)

    def courses_assignments_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.Sequence[lms.model.assignments.Assignment]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        return lms.backend.canvas.courses.assignments.list.request(self, parsed_course_id)

    def courses_users_fetch(self,
            course_id: str,
            user_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.users.CourseUser, None]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_user_id = lms.util.parse.required_int(user_id, 'user_id')
        return lms.backend.canvas.courses.users.fetch.request(self, parsed_course_id, parsed_user_id)

    def courses_users_list(self,
            course_id: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> typing.Sequence[lms.model.users.CourseUser]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        return lms.backend.canvas.courses.users.list.request(self, parsed_course_id)



"""
Supporting code for verifying test data.
See `lms.procedure.server` for the inputs (parser) to this procedure.
"""

import argparse
import typing

import edq.procedure.verify_exchanges

import lms.procedure.server

def run(args: typing.Union[argparse.Namespace, typing.Dict[str, typing.Any]]) -> int:
    """
    Run the procedure.
    The arguments may come directly from the parser for lms.cli.lib.verify-test-data.
    """

    if (not isinstance(args, dict)):
        args = vars(args)

    args.update(args.get('_config', {}))

    test_data_dir = args.get('test_data_dir', None)
    if (test_data_dir is None):
        raise ValueError("No test data dir was providded.")

    server_runner = lms.procedure.server.ServerRunner(**args)
    server_runner.start()

    failure_count = int(edq.procedure.verify_exchanges.run([test_data_dir], server_runner.server))

    server_runner.stop()

    return failure_count



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model.users

BASE_ENDPOINT = "/api/v1/courses/{course_id}/users?user_ids[]={user_id}"

def request(backend: typing.Any,
        course_id: int,
        user_id: int,
        include_role: bool = True,
        ) -> typing.Union[lms.backend.canvas.model.users.CourseUser, None]:
    """ Fetch a single course user. """

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id, user_id = user_id)
    headers = backend.get_standard_headers()

    if (include_role):
        url += '&include[]=enrollments'

    raw_object = lms.backend.canvas.common.make_get_request(url, headers, raise_on_404 = False)
    if (raw_object is None):
        return None

    if (len(raw_object) != 1):
        return None

    return lms.backend.canvas.model.users.CourseUser(**raw_object[0])



"""
Supporting code for running an LMS server for some larger process
(like generating or verifying test data).
"""

import argparse
import atexit
import logging
import signal
import subprocess
import typing

import edq.util.dirent
import edq.util.net
import edq.util.reflection

import lms.backend.backend
import lms.cli.parser
import lms.model.constants
import lms.util.net

DEFAULT_STARTUP_WAIT_SECS: float = 10.0
SERVER_STOP_WAIT_SECS: float = 5.00

class ServerRunner():
    """
    A class for running an LMS server for some sort of larger process.
    """

    def __init__(self,
            server: typing.Union[str, None] = None,
            backend_type: typing.Union[str, None] = None,
            server_start_command: typing.Union[str, None] = None,
            server_stop_command: typing.Union[str, None] = None,
            http_exchanges_out_dir: typing.Union[str, None] = None,
            server_output_path: typing.Union[str, None] = None,
            startup_wait_secs: typing.Union[float, None] = None,
            **kwargs: typing.Any) -> None:
        if (server is None):
            raise ValueError('No server specified.')

        self.server: str = server
        """ The server address to point requests to. """

        if (server_start_command is None):
            raise ValueError('No command to start the server was specified.')

        self.backend_type: typing.Union[str, None] = backend_type
        """
        The type of server being run.
        This value will be resolved after the server is started
        (since part of resolution may involve pinging the server.
        """

        self.server_start_command: str = server_start_command
        """ The server_start_command to run the LMS server. """

        self.server_stop_command: typing.Union[str, None] = server_stop_command
        """ An optional command to stop the server. """

        if (http_exchanges_out_dir is None):
            http_exchanges_out_dir = edq.util.dirent.get_temp_dir(prefix = 'edq-lms-http-exchanges-', rm = False)

        self.http_exchanges_out_dir: str = http_exchanges_out_dir
        """ Where to output the HTTP exchanges. """

        if (server_output_path is None):
            server_output_path = edq.util.dirent.get_temp_path(prefix = 'edq-lms-server-output-', rm = False) + '.txt'

        self.server_output_path: str = server_output_path
        """ Where to write server output (stdout and stderr). """

        if (startup_wait_secs is None):
            startup_wait_secs = DEFAULT_STARTUP_WAIT_SECS

        self.startup_wait_secs = startup_wait_secs
        """ How long to wait after the server start command is run before making requests to the server. """

        self._old_exchanges_out_dir: typing.Union[str, None] = None
        """
        The value of edq.util.net._exchanges_out_dir when start() is called.
        The original value may be changed in start(), and will be reset in stop().
        """

        self._old_exchanges_clean_func: typing.Union[str, None] = None
        """
        The value of edq.util.net._exchanges_clean_func when start() is called.
        The original value may be changed in start(), and will be reset in stop().
        """

        self._old_set_exchanges_clean_func: bool = False
        """
        The value of lms.util.net._set_exchanges_clean_func when start() is called.
        The original value may be changed in start(), and will be reset in stop().
        """

        self._process: typing.Union[subprocess.Popen, None] = None
        """ The server process. """

        self._server_output_file: typing.Union[typing.IO, None] = None
        """ The file that server output is written to. """

    def start(self) -> None:
        """ Start the server. """

        if (self._process is not None):
            return

        # Ensure stop() is called.
        atexit.register(self.stop)

        # Start the server.

        logging.info("Writing HTTP exchanges to '%s'.", self.http_exchanges_out_dir)
        logging.info("Writing server output to '%s'.", self.server_output_path)
        logging.info("Starting the server ('%s') and waiting %0.2f seconds.", self.server, self.startup_wait_secs)

        self._server_output_file = open(self.server_output_path, 'a', encoding = edq.util.dirent.DEFAULT_ENCODING)  # pylint: disable=consider-using-with
        self._process = subprocess.Popen(self.server_start_command,  # pylint: disable=consider-using-with
                shell = True, stdout = self._server_output_file, stderr = subprocess.STDOUT)

        status = None
        try:
            # Ensure the server is running cleanly.
            status = self._process.wait(self.startup_wait_secs)
        except subprocess.TimeoutExpired:
            # Good, the server is running.
            pass

        if (status is not None):
            hint = f"code: '{status}'"
            if (status == 125):
                hint = 'server may already be running'

            raise ValueError(f"Server was unable to start successfully ('{hint}').")

        # Resolve the backend type.
        self.backend_type = lms.backend.backend.guess_backend_type(self.server, backend_type = self.backend_type)
        if (self.backend_type is None):
            raise ValueError(f"Unable to determine backend type for server '{self.server}'.")

        # Store and set networking config.

        self._old_set_exchanges_clean_func = lms.cli.parser._set_exchanges_clean_func
        lms.cli.parser._set_exchanges_clean_func = False

        exchange_clean_func = lms.model.constants.BACKEND_REQUEST_CLEANING_FUNCS.get(self.backend_type, lms.util.net.clean_lms_response)
        exchange_clean_func_name = edq.util.reflection.get_qualified_name(exchange_clean_func)

        self._old_exchanges_out_dir = edq.util.net._exchanges_out_dir
        edq.util.net._exchanges_out_dir = self.http_exchanges_out_dir

        self._old_exchanges_clean_func = edq.util.net._exchanges_clean_func
        edq.util.net._exchanges_clean_func = exchange_clean_func_name

    def stop(self) -> None:
        """ Stop the server. """

        if (self._process is None):
            return

        # Restore networking config.

        lms.cli.parser._set_exchanges_clean_func = self._old_set_exchanges_clean_func
        self._old_set_exchanges_clean_func = False

        edq.util.net._exchanges_out_dir = self._old_exchanges_out_dir
        self._old_exchanges_out_dir = None

        edq.util.net._exchanges_clean_func = self._old_exchanges_clean_func
        self._old_exchanges_clean_func = None

        # Stop the server.
        logging.info('Stopping the server.')
        self._stop_server()

        self._process = None

        if (self._server_output_file is not None):
            self._server_output_file.close()
            self._server_output_file = None

    def _stop_server(self) -> typing.Union[int, None]:
        if (self._process is None):
            return None

        # Check if the process is already dead.
        status = self._process.poll()
        if (status is not None):
            return status

        # If the user provided a special command, try it.
        if (self.server_stop_command is not None):
            subprocess.run(self.server_stop_command,
                    shell = True, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL,
                    check = False)

        status = self._process.poll()
        if (status is not None):
            return status

        # Try to end the server gracefully.
        try:
            self._process.send_signal(signal.SIGINT)
            self._process.wait(SERVER_STOP_WAIT_SECS)
        except subprocess.TimeoutExpired:
            pass

        status = self._process.poll()
        if (status is not None):
            return status

        # End the server hard.
        try:
            self._process.kill()
            self._process.wait(SERVER_STOP_WAIT_SECS)
        except subprocess.TimeoutExpired:
            pass

        status = self._process.poll()
        if (status is not None):
            return status

        return None

def modify_parser(parser: argparse.ArgumentParser) -> None:
    """ Modify the parser to add arguments for running a server. """

    parser.add_argument('server_start_command', metavar = 'RUN_SERVER_COMMAND',
        action = 'store', type = str,
        help = 'The command to run the LMS server that will be the target of the data generation commands.')

    parser.add_argument('--startup-wait', dest = 'startup_wait_secs',
        action = 'store', type = float, default = DEFAULT_STARTUP_WAIT_SECS,
        help = 'The time to wait between starting the server and sending commands (default: %(default)s).')

    parser.add_argument('--server-output-file', dest = 'server_output_path',
        action = 'store', type = str, default = None,
        help = 'Where server output will be written. Defaults to a random temp file.')

    parser.add_argument('--server-stop-command', dest = 'server_stop_command',
        action = 'store', type = str, default = None,
        help = 'An optional command to stop the server. After this the server will be sent a SIGINT and then a SIGKILL.')



import typing

import lms.model.users

CORE_FIELDS: typing.List[str] = [
    'id',
    'email',
]

ENROLLMENT_TYPES: typing.List[str] = [
    'ObserverEnrollment',
    'StudentEnrollment',
    'TaEnrollment',
    'DesignerEnrollment',
    'TeacherEnrollment',
]
"""
Canvas enrollment types ordered by priority/power.
The higher the index, the more power.
"""

class CourseUser(lms.model.users.CourseUser):
    """
    A Canvas user associated with a course.

    Common fields will be held in lms.model.users.CourseUser.
    Fields that are not common, but used by this backend will be explicitly listed.
    Other fields coming from Canvas will be held in lms.model.users.CourseUser.extra_fields.

    See: https://developerdocs.instructure.com/services/canvas/resources/users
    """

    def __init__(self,
            enrollments: typing.Union[typing.Any, None] = None,
            **kwargs: typing.Any) -> None:
        # Check for important fields.
        for field in CORE_FIELDS:
            if (field not in kwargs):
                raise ValueError(f"Canvas user is missing '{field}'.")

        # Modify specific arguments before sending them to super.
        kwargs['id'] = str(kwargs['id'])
        kwargs['username'] = kwargs.get('login_id', None)

        if (enrollments is not None):
            kwargs['role'] = self._parse_role_from_enrollments(enrollments)

        super().__init__(**kwargs)

        if (enrollments is None):
            enrollments = []

        if (not isinstance(enrollments, list)):
            raise ValueError(f"Enrollments should be a list of enrollments, found {type(enrollments)}.")

        for (i, enrollment) in enumerate(enrollments):
            if (not isinstance(enrollment, dict)):
                raise ValueError(f"Enrollment at index {i} should be a dict, found {type(enrollment)}.")

        self.enrollments: typing.List[typing.Dict[str, typing.Any]] = enrollments
        """
        This field can be requested with certain API calls, and will return a list of the users active enrollments.
        See the List enrollments API for more details about the format of these records.
        """

    def _parse_role_from_enrollments(self, enrollments: typing.Any) -> typing.Union[str, None]:
        """
        Try to parse the user's role from their enrollments.
        If multiple roles are discovered, take the "highest" one.

        See: https://developerdocs.instructure.com/services/canvas/resources/enrollments
        """

        if (not isinstance(enrollments, list)):
            return None

        best_role = None
        best_index = -1

        for enrollment in enrollments:
            if (not isinstance(enrollment, dict)):
                continue

            if (enrollment.get('enrollment_state', None) != 'active'):
                continue

            role = enrollment.get('role', None)

            role_index = -1
            if (role in ENROLLMENT_TYPES):
                role_index = ENROLLMENT_TYPES.index(role)

            if ((best_role is None) or (role_index > best_index)):
                best_role = role
                best_index = role_index

        return best_role



import typing

import edq.util.time

import lms.model.base

class Assignment(lms.model.base.BaseType):
    """
    An assignment within a course.
    """

    CORE_FIELDS = [
        'id', 'name', 'description',
        'open_date', 'close_date', 'due_date',
        'points_possible', 'position', 'group_id',
    ]

    def __init__(self,
            id: typing.Union[str, None] = None,
            name: typing.Union[str, None] = None,
            description: typing.Union[str, None] = None,
            open_date: typing.Union[edq.util.time.Timestamp, None] = None,
            close_date: typing.Union[edq.util.time.Timestamp, None] = None,
            due_date: typing.Union[edq.util.time.Timestamp, None] = None,
            points_possible: typing.Union[float, None] = None,
            position: typing.Union[int, None] = None,
            group_id: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.id: typing.Union[str, None] = id
        """ The LMS's identifier for this assignment. """

        self.name: typing.Union[str, None] = name
        """ The display name of this assignment. """

        self.description: typing.Union[str, None] = description
        """ The description of this assignment. """

        self.open_date: typing.Union[edq.util.time.Timestamp, None] = open_date
        """ The datetime that this assignment becomes open at. """

        self.close_date: typing.Union[edq.util.time.Timestamp, None] = close_date
        """ The datetime that this assignment becomes close at. """

        self.due_date: typing.Union[edq.util.time.Timestamp, None] = due_date
        """ The datetime that this assignment is due at. """

        self.points_possible: typing.Union[float, None] = points_possible
        """ The maximum number of points possible for this assignment. """

        self.position: typing.Union[int, None] = position
        """ The order that this assignment should appear relative to other assignments. """

        self.group_id: typing.Union[str, None] = group_id
        """ The LMS's identifier for the group this assignment appears in. """

class AssignmentQuery():
    """
    A class for the different ways one can attempt to reference an LMS assignment.
    In general, an assignment can be queried by:
     - LMS Assignment ID (`id`)
     - Full Name (`name`)
     - f"{name} ({id})"
    """

    def __init__(self,
            id: typing.Union[str, None] = None,
            name: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        self.id: typing.Union[str, None] = id
        """ The LMS's identifier for this query. """

        self.name: typing.Union[str, None] = name
        """ The display name of this query. """

        if ((self.id is None) and (self.name is None)):
            raise ValueError("Assignment query is empty, it must have at least one piece of information (id, name).")

    def requires_resolution(self) -> bool:
        """
        Check if this query needs to be resolved.
        Typically, this means that the query is not just an LMS ID.
        """

        return ((self.id is None) or (self.name is not None))

    def match(self, target: Assignment) -> bool:
        """ Check if this query matches an assignment. """

        for field_name in ['id', 'name']:
            self_value = getattr(self, field_name, None)
            target_value = getattr(target, field_name, None)

            if (self_value is None):
                continue

            if (self_value != target_value):
                return False

        return True

    def __str__(self) -> str:
        text = self.name

        if (self.id is not None):
            if (text is not None):
                text = f"{text} ({id})"
            else:
                text = self.id

        if (text is None):
            return '<unknown>'

        return text



"""
The `lms.procedure` package contains code for more complex operations (referred to as "procedures").
These procedures are often processes that require several different components,
and this package provides a place put them.
"""



import re
import typing

import lms.model.assignments
import lms.model.users

T = typing.TypeVar('T')

class APIBackend():
    """
    API backends provide a unified interface to an LMS.

    Note that instead of using an abstract class,
    methods will raise a NotImplementedError by default.
    This will allow child backends to fill in as much functionality as they can,
    while still leaving gaps where they are incomplete or impossible.
    """

    def __init__(self,
            server: str,
            **kwargs: typing.Any) -> None:
        self.server: str = server
        """ The server this backend will connect to. """

    # API Methods

    def courses_assignments_get(self,
            course_id: str,
            queries: typing.List[lms.model.assignments.AssignmentQuery],
            **kwargs: typing.Any) -> typing.Sequence[lms.model.assignments.Assignment]:
        """
        Get the specified assignments associated with the given course.
        """

        if (len(queries) == 0):
            return []

        # Check if at least one of the queries requires resolution.
        # If resolution is required, then just list the assignments and match the queries.
        if (any(query.requires_resolution() for query in queries)):
            return self.courses_assignments_list_and_resolve(course_id, queries, **kwargs)

        # If there are multiple queries, then just list the assignments and match the queries.
        if (len(queries) > 1):
            return self.courses_assignments_list_and_resolve(course_id, queries, **kwargs)

        # If there is just one query, then fetch it.
        result = self.courses_assignments_fetch(course_id, typing.cast(str, queries[0].id), **kwargs)
        if (result is None):
            return []

        return [result]

    def courses_assignments_fetch(self,
            course_id: str,
            assignment_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.assignments.Assignment, None]:
        """
        Fetch a single assignment associated with the given course.
        Return None if no matching assignment is found.
        """

        raise NotImplementedError('courses_assignments_fetch')

    def courses_assignments_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.Sequence[lms.model.assignments.Assignment]:
        """
        Get the assignments associated with the given course.
        """

        raise NotImplementedError('courses_assignments_list')

    def courses_assignments_list_and_resolve(self,
            course_id: str,
            queries: typing.List[lms.model.assignments.AssignmentQuery],
            **kwargs: typing.Any) -> typing.Sequence[lms.model.assignments.Assignment]:
        """
        List the course assignments and then match the given queries.
        """

        assignments = self.courses_assignments_list(course_id, **kwargs)

        matches = []
        for assignment in assignments:
            for query in queries:
                if (query.match(assignment)):
                    matches.append(assignment)
                    break

        return matches

    def courses_users_get(self,
            course_id: str,
            queries: typing.List[lms.model.users.UserQuery],
            **kwargs: typing.Any) -> typing.Sequence[lms.model.users.CourseUser]:
        """
        Get the specified users associated with the given course.
        """

        if (len(queries) == 0):
            return []

        # Check if at least one of the queries requires resolution.
        # If resolution is required, then just list the users and match the queries.
        if (any(query.requires_resolution() for query in queries)):
            return self.courses_users_list_and_resolve(course_id, queries, **kwargs)

        # If there are multiple queries, then just list the users and match the queries.
        if (len(queries) > 1):
            return self.courses_users_list_and_resolve(course_id, queries, **kwargs)

        # If there is just one query, then fetch it.
        result = self.courses_users_fetch(course_id, typing.cast(str, queries[0].id), **kwargs)
        if (result is None):
            return []

        return [result]

    def courses_users_fetch(self,
            course_id: str,
            user_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.users.CourseUser, None]:
        """
        Fetch a single user associated with the given course.
        Return None if no matching user is found.
        """

        raise NotImplementedError('courses_users_fetch')

    def courses_users_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.Sequence[lms.model.users.CourseUser]:
        """
        Get the users associated with the given course.
        """

        raise NotImplementedError('courses_users_list')

    def courses_users_list_and_resolve(self,
            course_id: str,
            queries: typing.List[lms.model.users.UserQuery],
            **kwargs: typing.Any) -> typing.Sequence[lms.model.users.CourseUser]:
        """
        List the course users and then match the given queries.
        """

        users = self.courses_users_list(course_id, **kwargs)

        matches = []
        for user in users:
            for query in queries:
                if (query.match(user)):
                    matches.append(user)
                    break

        return matches

    # Utility Methods

    def parse_assignment_query(self, text: typing.Union[str, None]) -> typing.Union[lms.model.assignments.AssignmentQuery, None]:
        """
        Attempt to parse an assignment query from a string.
        The there is no query, return a None.
        If the query is malformed, raise an exception.

        By default, this method assumes that LMS IDs are ints.
        Child backends may override this to implement their specific behavior.
        """

        return self._parse_int_query(lms.model.assignments.AssignmentQuery, text, check_email = False)

    def parse_assignment_queries(self, texts: typing.List[typing.Union[str, None]]) -> typing.List[lms.model.assignments.AssignmentQuery]:
        """ Parse a list of assignment queries. """

        queries = []
        for text in texts:
            query = self.parse_assignment_query(text)
            if (query is not None):
                queries.append(query)

        return queries

    def parse_user_query(self, text: typing.Union[str, None]) -> typing.Union[lms.model.users.UserQuery, None]:
        """
        Attempt to parse a user query from a string.
        The there is no query, return a None.
        If the query is malformed, raise an exception.

        By default, this method assumes that LMS IDs are ints.
        Child backends may override this to implement their specific behavior.
        """

        return self._parse_int_query(lms.model.users.UserQuery, text, check_email = True)

    def parse_user_queries(self, texts: typing.List[typing.Union[str, None]]) -> typing.List[lms.model.users.UserQuery]:
        """ Parse a list of user queries. """

        queries = []
        for text in texts:
            query = self.parse_user_query(text)
            if (query is not None):
                queries.append(query)

        return queries

    def _parse_int_query(self, query_type: typing.Type[T], text: typing.Union[str, None],
            check_email: bool = True,
            ) -> typing.Union[T, None]:
        """
        Parse a query with the assumption that LMS ids are ints.

        Accepts queries are in the following forms:
         - LMS ID (`id`)
         - Email (`email`)
         - Full Name (`name`)
         - f"{email} ({id})"
         - f"{name} ({id})"
        """

        if (text is None):
            return None

        # Clean whitespace.
        text = re.sub(r'\s+', ' ', str(text)).strip()
        if (len(text) == 0):
            return None

        id = None
        email = None
        name = None

        match = re.search(r'^(\S.*)\((\d+)\)$', text)
        if (match is not None):
            # Query has both text and id.
            name = match.group(1).strip()
            id = match.group(2)
        elif (re.search(r'^\d+$', text) is not None):
            # Query must be an ID.
            id = text
        else:
            name = text

        # Check if the name is actually an email address.
        if (check_email and (name is not None) and ('@' in name)):
            email = name
            name = None

        data = {
            'id': id,
            'name': name,
            'email': email,
        }

        return query_type(**data)



"""
Utilities for network and HTTP.
"""

import json
import typing

import requests

import edq.util.json

CANVAS_CLEAN_REMOVE_CONTENT_KEYS: typing.List[str] = [
    'created_at',
    'last_activity_at',
    'lti_context_id',
    'secure_params',
    'total_activity_time',
    'updated_at',
]
""" Keys to remove from Canvas content. """

def clean_lms_response(response: requests.Response, body: str) -> str:
    """
    A ResponseModifierFunction that attempt to identify
    if the requests comes from a Learning Management System (LMS),
    and clean the response accordingly.
    """

    for key in response.headers:
        key = key.lower().strip()

        if ('canvas' in key):
            return clean_canvas_response(response, body)

    return body

def clean_canvas_response(response: requests.Response, body: str) -> str:
    """
    See clean_lms_response(), but specifically for the Canvas LMS.
    This function will:
     - Remove X- headers.
     - Remove content keys: [last_activity_at, total_activity_time]
    """

    for key in list(response.headers.keys()):
        if (key.strip().lower().startswith('x-')):
            response.headers.pop(key, None)

    # Most canvas responses are JSON.
    try:
        data = edq.util.json.loads(body)
    except json.JSONDecodeError:
        # Response is not JSON.
        return body

    # Remove any content keys.
    _recursive_remove_keys(data, set(CANVAS_CLEAN_REMOVE_CONTENT_KEYS))

    # Convert body back to a string.
    body = edq.util.json.dumps(data)

    return body

def _recursive_remove_keys(data: typing.Any, remove_keys: typing.Set[str]) -> None:
    """
    Recursively descend through the given and remove any instance to the given key from any dictionaries.
    The data should only be simple types (POD, dicts, lists, tuples).
    """

    if (isinstance(data, (list, tuple))):
        for item in data:
            _recursive_remove_keys(item, remove_keys)
    elif (isinstance(data, dict)):
        for key in list(data.keys()):
            if (key in remove_keys):
                del data[key]
            else:
                _recursive_remove_keys(data[key], remove_keys)



import lms.model.users

# {course_id: {email: user, ...}, ...}
COURSE_USERS = {
    '1': {
        'course-admin@test.edulinq.org': lms.model.users.CourseUser(
            email = 'course-admin@test.edulinq.org',
            id = '2',
            name = 'course-admin',
            username = 'course-admin@test.edulinq.org',
            role = 'TaEnrollment',
        ),
        'course-grader@test.edulinq.org': lms.model.users.CourseUser(
            email = 'course-grader@test.edulinq.org',
            id = '3',
            name = 'course-grader',
            username = 'course-grader@test.edulinq.org',
            role = 'TaEnrollment',
        ),
        'course-other@test.edulinq.org': lms.model.users.CourseUser(
            email = 'course-other@test.edulinq.org',
            id = '4',
            name = 'course-other',
            username = 'course-other@test.edulinq.org',
            role = 'ObserverEnrollment',
        ),
        'course-owner@test.edulinq.org': lms.model.users.CourseUser(
            email = 'course-owner@test.edulinq.org',
            id = '5',
            name = 'course-owner',
            username = 'course-owner@test.edulinq.org',
            role = 'TeacherEnrollment',
        ),
        'course-student@test.edulinq.org': lms.model.users.CourseUser(
            email = 'course-student@test.edulinq.org',
            id = '6',
            name = 'course-student',
            username = 'course-student@test.edulinq.org',
            role = 'StudentEnrollment',
        ),
    },
}



import lms.model.assignments

# {course_id: {name: assignment, ...}, ...}
COURSE_ASSIGNMENTS = {
    '1': {
        '1': lms.model.assignments.Assignment(
            id = '1',
            name = 'Homework 0',
            points_possible = 2.0,
            position = 1,
            group_id = '1',
        ),
    },
    '2': {
        '2': lms.model.assignments.Assignment(
            id = '2',
            name = 'A Simple Bash Assignment',
            points_possible = 10.0,
            position = 1,
            group_id = '2'
        ),
        '3': lms.model.assignments.Assignment(
            id = '3',
            name = 'A Simple C++ Assignment',
            points_possible = 10.0,
            position = 2,
            group_id = '2'
        ),
        '4': lms.model.assignments.Assignment(
            id = '4',
            name = 'A Simple Java Assignment',
            points_possible = 10.0,
            position = 3,
            group_id = '2'
        ),
    }
}



import lms.backend.testing
import lms.model.testdata.assignments

def test_courses_assignments_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing course assignments. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {},
            [
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['1']['1'],
            ],
            None,
        ),

        (
            {
                'course_id': '2',
            },
            [
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['2']['2'],
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['2']['3'],
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['2']['4'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_assignments_list, test_cases)



import typing

import lms.backend.canvas.backend
import lms.model.constants
import lms.model.backend

def get_backend(
        server: typing.Union[str, None] = None,
        backend_type: typing.Union[str, None] = None,
        **kwargs: typing.Any) -> lms.model.backend.APIBackend:
    """
    Get an instance of an API backend from the given information.
    If the backend type is not explicitly provided,
    this function will attempt to guess it from other information.
    """

    if (server is None):
        raise ValueError("No LMS server address provided.")

    server = server.strip()
    if (not server.startswith('http')):
        server = 'http://' + server

    backend_type = guess_backend_type(server, backend_type = backend_type)

    if (backend_type == lms.model.constants.BACKEND_TYPE_CANVAS):
        return lms.backend.canvas.backend.CanvasBackend(server, **kwargs)

    raise ValueError(f"Unknown backend type: '{backend_type}'. Known backend types: {lms.model.constants.BACKEND_TYPES}.")

def guess_backend_type(
        server: typing.Union[str, None] = None,
        backend_type: typing.Union[str, None] = None,
        ) -> str:
    """
    Attempt to guess the backend type from a server.
    This function will raise if it cannot guess the backend type.
    """

    if (backend_type is not None):
        return backend_type

    if (server is not None):
        if ('canvas' in server.lower()):
            return lms.model.constants.BACKEND_TYPE_CANVAS

    raise ValueError(f"Unable to guess backend type from server: '{server}'.")



"""
Supporting code for generating test data.
See `lms.procedure.server` for the inputs (parser) to this procedure.
"""

import argparse
import os
import typing

import edq.testing.run

import lms.backend.testing
import lms.procedure.server

THIS_DIR: str = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
ROOT_PACKAGE_DIR: str = os.path.join(THIS_DIR, '..')

def run(args: typing.Union[argparse.Namespace, typing.Dict[str, typing.Any]]) -> int:
    """
    Run the procedure.
    The arguments may come directly from the parser for lms.cli.lib.generate-test-data.
    """

    if (not isinstance(args, dict)):
        args = vars(args)

    args.update(args.get('_config', {}))

    server_runner = lms.procedure.server.ServerRunner(**args)
    server_runner.start()

    # Configure backend tests.
    lms.backend.testing.BackendTest.allowed_backend = server_runner.backend_type
    lms.backend.testing.BackendTest.skip_test_exchanges_base = True
    lms.backend.testing.BackendTest.override_server_url = server_runner.server

    # Run the tests (which generate the data).
    test_args = {
        'test_dirs': [ROOT_PACKAGE_DIR],
    }
    failure_count = int(edq.testing.run.run(test_args))

    server_runner.stop()

    return failure_count



import typing

import lms.backend.canvas.common
import lms.model.assignments
import lms.util.parse

CORE_FIELDS: typing.List[str] = [
    'id',
]

class Assignment(lms.model.assignments.Assignment):
    """
    A Canvas assignment associated with a course.

    Common fields will be held in lms.model.assignments.Assignment.
    Fields that are not common, but used by this backend will be explicitly listed.
    Other fields coming from Canvas will be held in lms.model.assignments.Assignment.extra_fields.

    See: https://developerdocs.instructure.com/services/canvas/resources/assignments
    """

    def __init__(self,
            **kwargs: typing.Any) -> None:
        # Check for important fields.
        for field in CORE_FIELDS:
            if (field not in kwargs):
                raise ValueError(f"Canvas assignment is missing '{field}'.")

        # Modify specific arguments before sending them to super.
        kwargs['id'] = lms.util.parse.required_string(kwargs['id'], 'id')
        kwargs['group_id'] = lms.util.parse.required_string(kwargs.get('assignment_group_id', None), 'id')
        kwargs['due_date'] = lms.backend.canvas.common.parse_timestamp(kwargs.get('due_at', None))
        kwargs['open_date'] = lms.backend.canvas.common.parse_timestamp(kwargs.get('unlock_at', None))
        kwargs['close_date'] = lms.backend.canvas.common.parse_timestamp(kwargs.get('lock_at', None))

        super().__init__(**kwargs)



import typing

import lms.model.base

class ServerUser(lms.model.base.BaseType):
    """
    A user associated with an LMS server.
    """

    CORE_FIELDS = ['id', 'name', 'email', 'username']
    """ The common fields shared across backends for this type. """

    def __init__(self,
            id: typing.Union[str, None] = None,
            email: typing.Union[str, None] = None,
            username: typing.Union[str, None] = None,
            name: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.id: typing.Union[str, None] = id
        """ The LMS's identifier for this user. """

        self.name: typing.Union[str, None] = name
        """ The display name of this user. """

        self.email: typing.Union[str, None] = email
        """ The email address of this user. """

        self.username: typing.Union[str, None] = username
        """ The username for this user (often overlaps with email). """

class CourseUser(ServerUser):
    """
    A user associated with a course, e.g., an instructor or student.
    """

    CORE_FIELDS = ServerUser.CORE_FIELDS + ['role']
    """ The common fields shared across backends for this type. """

    def __init__(self,
            role: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.role: typing.Union[str, None] = role
        """ The role of this user within this course (e.g., instructor, student). """

class UserQuery():
    """
    A class for the different ways one can attempt to reference an LMS user.
    In general, a user can be queried by:
     - LMS User ID (`id`)
     - Email (`email`)
     - Full Name (`name`)
     - f"{email} ({id})"
     - f"{name} ({id})"
    """

    def __init__(self,
            id: typing.Union[str, None] = None,
            name: typing.Union[str, None] = None,
            email: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        self.id: typing.Union[str, None] = id
        """ The LMS's identifier for this query. """

        self.name: typing.Union[str, None] = name
        """ The display name of this query. """

        self.email: typing.Union[str, None] = email
        """ The email address of this query. """

        if ((self.id is None) and (self.name is None) and (self.email is None)):
            raise ValueError("User query is empty, it must have at least one piece of information (id, email, name).")

    def requires_resolution(self) -> bool:
        """
        Check if this query needs to be resolved.
        Typically, this means that the query is not just an LMS ID.
        """

        return ((self.id is None) or (self.name is not None) or (self.email is not None))

    def match(self, target: ServerUser) -> bool:
        """ Check if this query matches a user. """

        for field_name in ['id', 'name', 'email']:
            self_value = getattr(self, field_name, None)
            target_value = getattr(target, field_name, None)

            if (self_value is None):
                continue

            if (self_value != target_value):
                return False

        return True

    def __str__(self) -> str:
        text = self.email
        if (text is None):
            text = self.name

        if (self.id is not None):
            if (text is not None):
                text = f"{text} ({id})"
            else:
                text = self.id

        if (text is None):
            return '<unknown>'

        return text



import typing

import lms.util.parse

def check_required_course(config: typing.Dict[str, typing.Any]) -> typing.Union[str, None]:
    """
    Fetch and ensure that a course is provided in the config.
    If no course is provided, print a message and return None.
    """

    course = lms.util.parse.optional_string(config.get('course', None))
    if (course is None):
        print('ERROR: No course has been provided.')

    return course
